// VncSharp - .NET VNC Client Library
// Copyright (C) 2004  David Humphrey, Chuck Borgh, Matt Cyr
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Threading;
using System.Drawing;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Windows.Forms;

using VncSharp.Encodings;

namespace VncSharp
{
	/// <summary>
	/// Delegate definition of an Event Handler used to indicate a Framebuffer Update has been received.
	/// </summary>
	public delegate void VncUpdateHandler(object sender, VncEventArgs e);
	
	public class VncClient
	{
		RfbProtocol rfb;				// The protocol object handling all communication with server.
		Framebuffer buffer;			// The geometry and properties of the remote framebuffer
		byte securityType;			// The type of Security agreed upon by client/server
		EncodedRectangleFactory factory;
		Thread worker;				// To request and read in-coming updates from server
		ManualResetEvent done;		// Used to tell the worker thread to die cleanly

		/// <summary>
		/// Raised when the connection to the remote host is lost.
		/// </summary>
		public event EventHandler ConnectionLost;
	
		public VncClient()
		{
		}

		/// <summary>
		/// Gets the Framebuffer representing the remote server's desktop geometry.
		/// </summary>
		public Framebuffer Framebuffer {
			get { 
				return buffer; 
			}
		}

		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <param name="display">The Display number (used on Unix hosts).</param>
		/// <param name="port">The Port number used by the Host, usually 5900.</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host, int display, int port)
		{
			if (host == null) throw new ArgumentNullException("host");

			// If a diplay number is specified (used to connect to Unix servers)
			// it must be 0 or greater.  This gets added to the default port number
			// in order to determine where the server will be listening for connections.
			if (display < 0) throw new ArgumentOutOfRangeException("display", display, "Display number must be non-negative.");
			port += display;
			
			rfb = new RfbProtocol();
			
			// Connect and determine version of server, and set client protocol version to match			
			try {
				rfb.Connect(host, port);
				rfb.ReadProtocolVersion();
				rfb.WriteProtocolVersion();

				// Figure out which type of authentication the server uses
				byte[] types = rfb.ReadSecurityTypes();
				
				// Based on what the server sends back in the way of supported Security Types, one of
				// two things will need to be done: either the server will reject the connection (i.e., type = 0),
				// or a list of supported types will be sent, of which we need to choose and use one.
				if (types.Length > 0) {
					if (types[0] == 0) {
						// The server is not able (or willing) to accept the connection.
						// A message follows indicating why the connection was dropped.
						throw new VncProtocolException("Connection Failed. The server rejected the connection for the following reason: " + rfb.ReadSecurityFailureReason());
					} else {
						securityType = GetSupportedSecurityType(types);
						Debug.Assert(securityType > 0, "Unknown Security Type(s)", "The server sent one or more unknown Security Types.");
						
						rfb.WriteSecurityType(securityType);
						
						// Protocol 3.8 states that a SecurityResult is still sent when using NONE (see 6.2.1)
						if (rfb.ServerVersion == 3.8f && securityType == 0) {
							if (rfb.ReadSecurityResult() > 0) {
								// For some reason, the server is not accepting the connection.  Get the
								// reason and throw and exception
								throw new VncProtocolException("Unable to Connecto to the Server. The Server rejected the connection for the following reason: " + rfb.ReadSecurityFailureReason());
							}
						}
						
						return (securityType > 1) ? true : false;
					}
				} else {
					// Something is wrong, since we should have gotten at least 1 Security Type
					throw new VncProtocolException("Protocol Error Connecting to Server. The Server didn't send any Security Types during the initial handshake.");
				}
			} catch (Exception e) {
				throw new VncProtocolException("Unable to connect to the server. Error was: " + e.Message, e);
			}			
		}

		#region	Connect() method Overloads
		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required. Default Display and Port numbers are used.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host)
		{
			return Connect(host, 0, 5900);
		}

		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required. The Port number is calculated based on the Display.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <param name="display">The Display number (used on Unix hosts).</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host, int display)
		{
			return Connect(host, display, 5900);
		}
		#endregion		

		/// <summary>
		/// Examines a list of Security Types supported by a VNC Server and chooses one that the Client supports.  See 6.1.2 of the RFB Protocol document v. 3.8.
		/// </summary>
		/// <param name="types">An array of bytes representing the Security Types supported by the VNC Server.</param>
		/// <returns>A byte that represents the Security Type to be used by the Client.</returns>
		protected byte GetSupportedSecurityType(byte[] types)
		{
			// Pick the first match in the list of given types.  If you want to add support for new
			// security types, do it here:
			for (int i = 0; i < types.Length; ++i) {
				if (   types[i] == 1  	// None
					|| types[i] == 2	// VNC Authentication
// None of the following are currently supported -------------------
//					|| types[i] == 5	// RA2
//					|| types[i] == 6    // RA2ne
//					|| types[i] == 16   // Tight
//					|| types[i] == 17 	// Ultra
//					|| types[i] == 18 	// TLS
				   ) return types[i];
			}
			return 0;
		}

		/// <summary>
		/// Use a password to authenticate with a VNC Host. NOTE: This is only necessary if Connect() returns TRUE.
		/// </summary>
		/// <param name="password">The password to use.</param>
		/// <returns>Returns True if Authentication worked, otherwise False.</returns>
		public bool Authenticate(string password)
		{
			if (password == null) throw new ArgumentNullException("password");
			
			// If new Security Types are supported in future, add the code here.  For now, only 
			// VNC Authentication is supported.
			if (securityType == 2) {
				PerformVncAuthentication(password);
			} else {
				throw new NotSupportedException("Unable to Authenticate with Server. The Server uses an Authentication scheme unknown to the client.");
			}
			
			if (rfb.ReadSecurityResult() == 0) {
				return true;
			} else {
				// Authentication failed, and if the server is using Protocol version 3.8, a 
				// plain text message follows indicating why the error happend.  I'm not 
				// currently using this message, but it is read here to clean out the stream.
				// In earlier versions of the protocol, the server will just drop the connection.
				if (rfb.ServerVersion == 3.8) rfb.ReadSecurityFailureReason();
				rfb.Close();	// TODO: Is this the right place for this???
				return false;
			}
		}

		/// <summary>
		/// Performs VNC Authentication using VNC DES encryption.  See the RFB Protocol doc 6.2.2.
		/// </summary>
		/// <param name="password">A string containing the user's password in clear text format.</param>
		protected void PerformVncAuthentication(string password)
		{
			byte[] challenge = rfb.ReadSecurityChallenge();
			rfb.WriteSecurityResponse(EncryptChallenge(password, challenge));
		}

		/// <summary>
		/// Encrypts a challenge using the specified password. See RFB Protocol Document v. 3.8 section 6.2.2.
		/// </summary>
		/// <param name="password">The user's password.</param>
		/// <param name="challenge">The challenge sent by the server.</param>
		/// <returns>Returns the encrypted challenge.</returns>
		protected byte[] EncryptChallenge(string password, byte[] challenge)
		{
			byte[] key = new byte[8];

			// Key limited to 8 bytes max.
			if (password.Length >= 8) {
				System.Text.Encoding.ASCII.GetBytes(password, 0, 8, key, 0);
			} else {
				System.Text.Encoding.ASCII.GetBytes(password, 0, password.Length, key, 0);
			}			

			// HACK: VNC reverses key order!?
			key = FixBug(key); 

			// VNC uses DES, not 3DES as written in some documentation
			DES des = new DESCryptoServiceProvider();
			des.FeedbackSize = 0;
			des.Padding = PaddingMode.None;
			des.Mode = CipherMode.ECB;

			ICryptoTransform enc = des.CreateEncryptor(key, null); 

			byte[] response = new byte[16];
			enc.TransformBlock(challenge, 0, challenge.Length, response, 0);
			
			return response;
		}

		/// <summary>VNC DES authentication has a bug, such that keys are reversed.  This code 
		/// was written by Dominic Ullmann (dominic_ullmann@swissonline.ch) and is 
		/// is being used under the GPL.</summary>
		/// <param name="desKey">The key to be altered.</param>
		/// <returns>Returns the fixed key as an array of bytes.</returns>
		protected byte[] FixBug(byte[] desKey)
		{
			byte[] newkey = new byte[8];

			for (int i = 0; i < 8; i++) {
				// revert desKey[i]:
				newkey[i] = (byte) (	
					((desKey[i] & 0x01) << 7) |  
					((desKey[i] & 0x02) << 5) |
					((desKey[i] & 0x04) << 3) |
					((desKey[i] & 0x08) << 1) |
					((desKey[i] & 0x10) >> 1) |
					((desKey[i] & 0x20) >> 3) |
					((desKey[i] & 0x40) >> 5) |
					((desKey[i] & 0x80) >> 7)
					);
			}
			return newkey;
		}
		
		/// <summary>
		/// Finish setting-up protocol with VNC Host.  Should be called after Connect and Authenticate (if password required).
		/// </summary>
		public void Initialize()
		{
			// Finish initializing protocol with host
			rfb.WriteClientInitialisation(false);
			buffer = rfb.ReadServerInit();
			rfb.WriteSetPixelFormat(buffer);	// just use the server's framebuffer format

			// TODO: currently Zrle is NOT supported...., 16});
//			rfb.WriteSetEncodings(new uint[] {5, 4, 2, 1, 0}); 
			rfb.WriteSetEncodings(new uint[] {	RfbProtocol.HEXTILE_ENCODING, 
											//	RfbProtocol.CORRE_ENCODING, // CoRRE is buggy in some hosts, so don't bother using
												RfbProtocol.RRE_ENCODING,
												RfbProtocol.COPYRECT_ENCODING,
												RfbProtocol.RAW_ENCODING });
			
			// Create an EncodedRectangleFactory so that EncodedRectangles can be built according to set pixel layout
			factory = new EncodedRectangleFactory(rfb, buffer);
		}

		/// <summary>
		/// Begin getting updates from the VNC Server.  This will continue until StopUpdates() is called.  NOTE: this must be called after Connect().
		/// </summary>
		public void StartUpdates()
		{
			// Start getting updates on background thread.
			worker = new Thread(new ThreadStart(this.GetRfbUpdates));
			worker.IsBackground = true;
			done = new ManualResetEvent(false);
			worker.Start();
		}

		/// <summary>
		/// Stops sending requests for updates and disconnects from the remote host.  You must call Connect() again if you wish to re-establish a connection.
		/// </summary>
		public void Disconnect()
		{
			// Stop the worker thread.
			done.Set();
			worker.Join();

			rfb.Close();	
			rfb = null;
		}

		/// <summary>
		/// An event that occurs whenever the server sends a Framebuffer Update.
		/// </summary>
		public event VncUpdateHandler VncUpdate;
		
		
		private bool CheckIfThreadDone()
		{
			return done.WaitOne(0, false);
		}
		
		/// <summary>
		/// Worker thread lives here and processes protocol messages infinitely, triggering events or other actions as necessary.
		/// </summary>
		private void GetRfbUpdates()
		{
			int rectangles;
			int enc;

			// Get the initial destkop from the host
			RequestScreenUpdate(true);

			while (true) {
				// See if thread should die
				if (CheckIfThreadDone()) {
					break;
//				} else {
//					Thread.Sleep(1); // necessary?
				}
				
				try {
					switch (rfb.ReadServerMessageType()) {
						case RfbProtocol.FRAMEBUFFER_UPDATE:
							rectangles = rfb.ReadFramebufferUpdate();

							if (CheckIfThreadDone()) {
								break;
							}

							for (int i = 0; i < rectangles; ++i) {
								// Get the update rectangle's info
								Rectangle rectangle;
								rfb.ReadFramebufferUpdateRectHeader(out rectangle, out enc);

								// Build a derived EncodedRectangle type and pull-down all the pixel info
								EncodedRectangle er = factory.Build(rectangle, enc);
								er.Decode();

								// Let the UI know that an updated rectangle is available, but check
								// to see if the user closed things down first.
								if (!CheckIfThreadDone() && VncUpdate != null) {
									VncEventArgs e = new VncEventArgs(er);

									// In order to play nicely with WinForms controls, we do a check here to 
									// see if it is necessary to synchronize this event with the UI thread.
									if (VncUpdate.Target is System.Windows.Forms.Control) {
										Control target = VncUpdate.Target as Control;
										
										if (target != null) {
											target.Invoke(VncUpdate, new object[] {this, e});
										}
									} else {
										// Target is not a WinForms control, so do it on this thread...
										VncUpdate(this, new VncEventArgs(er));
									}
								}
							}
							break;
						case RfbProtocol.BELL:
							Beep(500, 300);  // TODO: are there better values than these?
							break;
						case RfbProtocol.SERVER_CUT_TEXT:
							if (CheckIfThreadDone()) {
								break;
							}
							Clipboard.SetDataObject(rfb.ReadServerCutText().Replace("\n", Environment.NewLine), true);
							break;
						case RfbProtocol.SET_COLOUR_MAP_ENTRIES:
	// TODO: Needs to be implemented fully
	//						rfb.ReadColourMapEntry();
							break;
					}
				} catch (Exception e) {
					System.Diagnostics.Debug.Assert(false, "Exception in worker thread", e.Message);
					OnConnectionLost();
				}
			}
		}

		protected void OnConnectionLost()
		{
			if (ConnectionLost != null) {
				ConnectionLost(this, EventArgs.Empty);
			}
		}
	
// There is no managed way to get a system beep (until Framework v.2.0). So depending on the platform, something external has to be called.
#if Win32
		[System.Runtime.InteropServices.DllImport("kernel32.dll")]
		private static extern bool Beep(int freq, int duration);
#else
		private bool Beep(int freq, int duration)	// bool just so it matches the Win32 API signature
		{
			// TODO: How to do this under Unix?
			System.Console.Write("Beep!");
			return true;
		}
#endif

		// TODO: This needs to be pushed into the protocol rather than expecting keysym from the caller.
		public void WriteKeyboardEvent(uint keysym, bool pressed)
		{
			rfb.WriteKeyEvent(keysym, pressed);
		}

		/// <summary>
		/// Translates a Windows Forms KeyEventArgs object into the appropriate VNC KeySym value,
		/// which can then be sent to a VNC Host as part of a KeyboardEvent (see the WriteKeyboardEvent method).
		/// </summary>
		/// <param name="e">The KeyEventArgs to be translated into a keysym value.  Typically this will come from
		/// an OnKeyUp, OnKeyDown, or OnKeyPress event in a Windows Forms control.</param>
		/// <returns></returns>
		public uint TranslateKey(System.Windows.Forms.KeyEventArgs e)
		{
			uint keycode = (uint)e.KeyCode;
			switch(e.KeyCode) {
				case Keys.Back:
					keycode = 0xff08;
					break;
				case Keys.Tab:
					keycode = 0xff09;
					break;
				case Keys.Return:
					keycode = 0xff0d;
					break;
				case Keys.Escape:
					keycode = 0xff1b;
					break;
				case Keys.Insert:
					keycode = 0xff63;
					break;
				case Keys.Delete:
					keycode = 0xffff;
					break;
				case Keys.Home:
					keycode = 0xff50;
					break;
				case Keys.End:
					keycode = 0xff57;
					break;
				case Keys.PageUp:
					keycode = 0xff55;
					break;
				case Keys.PageDown:
					keycode = 0xff56;
					break;
				case Keys.Left:
					keycode = 0xff51;
					break;
				case Keys.Up:
					keycode = 0xff52;
					break;
				case Keys.Right:
					keycode = 0xff53;
					break;
				case Keys.Down:
					keycode = 0xff54;
					break;
				case Keys.F1:
					keycode = 0xffbe;
					break;
				case Keys.F2:
					keycode = 0xffbf;
					break;
				case Keys.F3:
					keycode = 0xffc0;
					break;
				case Keys.F4:
					keycode = 0xffc1;
					break;
				case Keys.F5:
					keycode = 0xffc2;
					break;
				case Keys.F6:
					keycode = 0xffc3;
					break;
				case Keys.F7:
					keycode = 0xffc4;
					break;
				case Keys.F8:
					keycode = 0xffc5;
					break;
				case Keys.F9:
					keycode = 0xffc6;
					break;
				case Keys.F10:
					keycode = 0xffc7;
					break;
				case Keys.F11:
					keycode = 0xffc8;
					break;
				case Keys.F12:
					keycode = 0xffc9;
					break;
				default:
					break;
			}

			// split these into seperate if statements just 
			// to know which keys are being manipulated
			if (keycode > 64 && keycode <= 90) {
				// A-Z
				if (!e.Shift) keycode += 32;
			} else if (keycode > 47 && keycode <= 57) { 
				//0-9 (well...1-0)
				if (e.Shift) {
					switch (keycode) {
						case 48:  // 0
							keycode -= 7;
							break;
						case 50:  // 2
							keycode += 14;
							break;
						case 49:  // 1
						case 51:  // 3
						case 52:  // 4
						case 53:  // 5
							keycode -= 16;
							break;
						case 54:  // 6
							keycode += 40;
							break;
						case 55:  // 7
						case 57:  // 9
							keycode -= 17;
							break;
						case 56:  // 8
							keycode -= 14;
							break;
					}
				}
			} else {
				switch (keycode) {
					case 192: // `
						keycode = 96;
						if (e.Shift) keycode += 30; // 126;
						break;
					case 188: // ,
						keycode = 44;
						if (e.Shift) keycode += 16;
						break;
					case 190: // .
						keycode = 46;
						if (e.Shift) keycode += 16;
						break;
					case 191: // /
						keycode = 47;
						if (e.Shift) keycode += 16;
						break;
					case 186: // ;
						keycode = 59;
						if (e.Shift) keycode -= 1;
						break;
					case 222: // '
						keycode = 39;
						if (e.Shift) keycode -= 5;
						break;
					case 219: // [
						keycode = 91;
						if (e.Shift) keycode += 32;
						break;
					case 221: // ]
						keycode = 93;
						if (e.Shift) keycode += 32;
						break;
					case 189: // -
						keycode = 45;
						if (e.Shift) keycode += 50;
						break;
					case 187: // =
						keycode = 61;
						if (e.Shift) keycode -= 18;
						break;
					case 220: // \
						keycode = 92;
						if (e.Shift) keycode += 32;
						break;
					default:
						break;
				}
			}
			return keycode;
		}

		// TODO: This needs to be pushed into the protocol rather than expecting the caller to create the mask.
		public void WritePointerEvent(byte buttonMask, Point point)
		{
			rfb.WritePointerEvent(buttonMask, point);
		}
		
		/// <summary>
		/// Requests that the remote host send a screen update.
		/// </summary>
		/// <param name="refreshFullScreen">TRUE if the entire screen should be refreshed, FALSE if only a partial region needs updating.</param>
		/// <remarks>RequestScreenUpdate needs to be called whenever the client screen needs to be updated to reflect the state of the remote 
		///	desktop.  Typically you only need to have a particular region of the screen updated and can still use the rest of the 
		/// pixels on the client-side (i.e., when moving the mouse pointer, only the area around the pointer changes).  Therefore, you should
		/// almost always set refreshFullScreen to FALSE.  If the client-side image becomes corrupted, call RequestScreenUpdate with
		/// refreshFullScreen set to TRUE to get the complete image sent again.
		/// </remarks>
		public void RequestScreenUpdate(bool refreshFullScreen)
		{
			rfb.WriteFramebufferUpdateRequest(0, 0, (ushort) buffer.Width, (ushort) buffer.Height, !refreshFullScreen);
		}
		
	}
}